try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version  # type: ignore

from schwifty.bic import BIC
from schwifty.iban import IBAN


__all__ = ["IBAN", "BIC"]
__version__ = version(__name__)



from __future__ import annotations

import re
import warnings
from typing import List
from typing import Optional

import iso3166
from pycountry import countries
from pycountry.db import Data

from schwifty import common
from schwifty import exceptions
from schwifty import registry


_bic_re = re.compile(r"[A-Z]{4}[A-Z]{2}[A-Z0-9]{2}(?:[A-Z0-9]{3})?")


class BIC(common.Base):
    """The BIC object.

    Examples:

        You can either create a new BIC object by providing a code as text::

            >>> bic = BIC('GENODEM1GLS')
            >>> bic.country_code
            'DE'
            >>> bic.location_code
            'M1'
            >>> bic.bank_code
            'GENO'

        or by using the :meth:`from_bank_code` classmethod::

            >>> bic = BIC.from_bank_code('DE', '43060967')
            >>> bic.formatted
            'GENO DE M1 GLS'

    Args:
        bic (str): The BIC number.
        allow_invalid (bool): If set to ``True`` validation is skipped on instantiation.

    Raises:
        InvalidLength: If the BIC's length is not 8 or 11 characters long.
        InvalidStructure: If the BIC contains unexpected characters.
        InvalidCountryCode: If the BIC's country code is unknown.
    """

    def __init__(self, bic: str, allow_invalid: Optional[bool] = False) -> None:
        super().__init__(bic)
        if not allow_invalid:
            self.validate()

    @classmethod
    def from_bank_code(cls, country_code: str, bank_code: str) -> BIC:
        """Create a new BIC object from country-code and domestic bank-code.

        Examples:
            >>> bic = BIC.from_bank_code('DE', '20070000')
            >>> bic.country_code
            'DE'
            >>> bic.bank_code
            'DEUT'
            >>> bic.location_code
            'HH'

            >>> BIC.from_bank_code('DE', '01010101')
            Traceback (most recent call last):
            ...
            InvalidBankCode: Unknown bank code '01010101' for country 'DE'


        Args:
            country_code (str): ISO 3166 alpha2 country-code.
            bank_code (str): Country specific bank-code.

        Returns:
            BIC: a BIC object generated from the given country code and bank code.

        Raises:
            InvalidBankCode: If the given bank code wasn't found in the registry

        Note:
            This currently only works for selected countries. Amongst them

            * Austria
            * Belgium
            * Croatia
            * Czech Republic
            * Finland
            * France
            * Germany
            * Great Britan
            * Latvia
            * Lithuania
            * Netherlands
            * Poland
            * Slovenia
            * Spain
            * Sweden
            * Switzerland
        """
        try:
            spec = registry.get("bank_code")
            assert isinstance(spec, dict)
            return cls(spec[(country_code, bank_code)]["bic"])
        except KeyError:
            raise exceptions.InvalidBankCode(
                f"Unknown bank code {bank_code!r} for country {country_code!r}"
            )

    def validate(self) -> bool:
        """Validate the structural integrity of this BIC.

        This function will verify the correct length, structure and the existence of the country
        code.

        Note:
            You have to use the `allow_invalid` paramter when constructing the :class:`BIC`-object
            to circumvent the implicit validation.

        Raises:
            InvalidLength: If the BIC's length is not 8 or 11 characters long.
            InvalidStructure: If the BIC contains unexpected characters.
            InvalidCountryCode: If the BIC's country code is unknown.
        """
        self._validate_length()
        self._validate_structure()
        self._validate_country_code()
        return True

    def _validate_length(self) -> None:
        if self.length not in (8, 11):
            raise exceptions.InvalidLength(f"Invalid length '{self.length}'")

    def _validate_structure(self) -> None:
        if not _bic_re.match(self.compact):
            raise exceptions.InvalidStructure(f"Invalid structure '{self.compact}'")

    def _validate_country_code(self) -> None:
        country_code = self.country_code
        try:
            iso3166.countries_by_alpha2[country_code]
        except KeyError:
            raise exceptions.InvalidCountryCode(f"Invalid country code '{country_code}'")

    @property
    def is_valid(self) -> bool:
        """bool: Indicate if this is a valid BIC.

        Note:
            You have to use the `allow_invalid` paramter when constructing the :class:`BIC`-object
            to circumvent the implicit validation.

        Examples:
            >>> BIC('FOOBARBAZ', allow_invalid=True).is_valid
            False

        .. versionadded:: 2020.08.1
        """
        try:
            return self.validate()
        except exceptions.SchwiftyException:
            return False

    @property
    def formatted(self) -> str:
        """str: The BIC separated in the blocks bank-, country- and location-code.

        Examples:
            >>> BIC('MARKDEF1100').formatted
            'MARK DE F1 100'
        """
        formatted = " ".join([self.bank_code, self.country_code, self.location_code])
        if self.branch_code:
            formatted += " " + self.branch_code
        return formatted

    def _lookup_values(self, key: str) -> List:
        spec = registry.get("bic")
        assert isinstance(spec, dict)
        entries = spec.get(self.compact, [])
        return sorted({entry[key] for entry in entries})

    @property
    def domestic_bank_codes(self) -> List[str]:
        """List[str]: The country specific bank-codes associated with the BIC.

        Examples:
            >>> BIC('MARKDEF1100').domestic_bank_codes
            ['10000000']

        .. versionadded:: 2020.01.0
        """
        return self._lookup_values("bank_code")

    @property
    def bank_names(self) -> List[str]:
        """List[str]: The name of the banks associated with the BIC.

        Examples:
            >>> BIC('MARKDEF1100').bank_names
            ['Bundesbank']

        .. versionadded:: 2020.01.0
        """
        return self._lookup_values("name")

    @property
    def bank_short_names(self) -> List[str]:
        """List[str]: The short name of the banks associated with the BIC.

        Examples:
            >>> BIC('MARKDEF1100').bank_short_names
            ['BBk Berlin']

        .. versionadded:: 2020.01.0
        """
        return self._lookup_values("short_name")

    @property
    def country_bank_code(self) -> Optional[str]:
        """str or None: The country specific bank-code associated with the BIC.

        .. deprecated:: 2020.01.0
           Use :meth:`domestic_bank_codes` instead.
        """
        warnings.warn("Use `BIC.domestic_bank_codes` instead", DeprecationWarning)
        codes = self.domestic_bank_codes
        return codes[0] if codes else None

    @property
    def bank_name(self) -> Optional[str]:
        """str or None: The name of the bank associated with the BIC.

        .. deprecated:: 2020.01.0
           Use :meth:`bank_names` instead.
        """
        warnings.warn("Use `BIC.bank_names` instead", DeprecationWarning)
        names = self.bank_names
        return names[0] if names else None

    @property
    def bank_short_name(self) -> Optional[str]:
        """str or None: The short name of the bank associated with the BIC.

        .. deprecated:: 2020.01.0
           Use :meth:`bank_short_names` instead.
        """
        warnings.warn("Use `BIC.bank_short_names` instead", DeprecationWarning)
        names = self.bank_short_names
        return names[0] if names else None

    @property
    def exists(self) -> bool:
        """bool: Indicates if the BIC is available in Schwifty's registry."""
        spec = registry.get("bic")
        assert isinstance(spec, dict)
        return bool(spec.get(self.compact))

    @property
    def type(self) -> str:
        """Indicates the type of BIC.

        This can be one of 'testing', 'passive', 'reverse billing' or 'default'

        Examples:
            >>> BIC('MARKDEF1100').type
            'passive'

        Returns:
            str: The BIC type.
        """
        if self.location_code[1] == "0":
            return "testing"
        elif self.location_code[1] == "1":
            return "passive"
        elif self.location_code[1] == "2":
            return "reverse billing"
        else:
            return "default"

    @property
    def country(self) -> Optional[Data]:
        """Country: The country this BIC is registered in."""
        return countries.get(alpha_2=self.country_code)

    @property
    def bank_code(self) -> str:
        """str: The bank-code part of the BIC."""
        return self._get_component(start=0, end=4)

    @property
    def country_code(self) -> str:
        """str: The ISO 3166 alpha2 country-code."""
        return self._get_component(start=4, end=6)

    @property
    def location_code(self) -> str:
        """str: The location code of the BIC."""
        return self._get_component(start=6, end=8)

    @property
    def branch_code(self) -> str:
        """str: The branch-code part of the BIC (if available)"""
        return self._get_component(start=8, end=11)


registry.build_index("bank", "bic", key="bic", accumulate=True)
registry.build_index("bank", "bank_code", key=("country_code", "bank_code"), primary=True)



import json

import requests


URL = "https://ewib.nbp.pl/plewibnra?dokNazwa=plewibnra.txt"


bic_remapping = {
    "PLUPLPRXXX": "POLUPLPRXXX",
    "POLUPLPRXX": "POLUPLPRXXX",
}


def process():
    registry = []
    with requests.get(URL, stream=True) as fp:
        for row in fp.iter_lines():
            cells = [cell.strip() for cell in row.decode("CP852").split("\t")]
            if len(cells) != 32:
                continue
            bic = cells[19].upper()
            registry.append(
                {
                    "country_code": "PL",
                    "primary": True,
                    "bic": bic_remapping.get(bic, bic),
                    "bank_code": cells[4],
                    "name": cells[1],
                    "short_name": cells[1],
                }
            )
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_pl.json", "w") as fp:
        json.dump(process(), fp, indent=2)



import csv
import json

import requests


URL = "https://www.cnb.cz/cs/platebni-styk/.galleries/ucty_kody_bank/download/kody_bank_CR.csv"


def process():
    with requests.get(URL, stream=True) as fp:
        csvfile = csv.reader([line.decode("latin1") for line in fp.iter_lines()], delimiter=";")
    return [
        {
            "country_code": "CZ",
            "primary": True,
            "bic": row[2].strip().upper(),
            "bank_code": row[0].strip(),
            "name": row[1].strip(),
            "short_name": row[1].strip(),
        }
        for i, row in enumerate(csvfile)
        if i >= 1 and row
    ]


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_cz.json", "w") as fp:
        json.dump(process(), fp, indent=2)



#!/usr/bin/env python
import json

import requests
import xlrd


# https://www.swedishbankers.se/fraagor-vi-arbetar-med/betalningar/ny-nordisk-betalningsinfrastruktur/iban-och-svenskt-nationellt-kontonummer/
URL = (
    "https://www.swedishbankers.se/media/4863/"
    "kalkylblad-i-iban-och-svenskt-nationellt-kontonummer-2021-02-16.xlsx"
)


def process():
    registry = []

    book = xlrd.open_workbook(file_contents=requests.get(URL).content)
    sheet = book.sheet_by_index(0)

    for row in list(sheet.get_rows())[3:]:
        bank_code, bic, name = row[:3]
        registry.append(
            {
                "country_code": "SE",
                "primary": True,
                "bic": bic.value.upper(),
                "bank_code": str(bank_code.value).split(".", 1)[0],
                "name": name.value.strip(),
                "short_name": name.value.strip(),
            }
        )
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_se.json", "w") as fp:
        json.dump(process(), fp, indent=2)



"""Checksum algorithms for German Bank institutes.

A description of the algorithms can be found on the website of the Bundesbank

https://www.bundesbank.de/resource/blob/603320/16a80c739bbbae592ca575905975c2d0/mL/pruefzifferberechnungsmethoden-data.pdf
"""
from dataclasses import dataclass
from functools import partial
from itertools import cycle
from typing import List
from typing import Optional

from schwifty import checksum
from schwifty.exceptions import InvalidBBANChecksum


ACCOUNT_CODE_LENGTH = 10

register = partial(checksum.register, prefix="DE")


@dataclass
class Positions:
    start: int
    end: int
    check_digit: int


def digit_sum(number: int) -> int:
    return sum(int(d) for d in str(number))


class WeightedModulus(checksum.Algorithm):
    minuend: Optional[int] = None
    modulus: int
    positions: Positions
    reverse: bool = True
    weights: List[int]

    def __init__(self) -> None:
        self.weighted_sum: int = 0
        self.remainder: int = 0

    def compute(self, account_code: str) -> str:
        digits = self.get_digits(self.adjust_input(account_code))
        self.remainder = self.compute_remainder(self.compute_weighted_sum(digits))
        if self.minuend is None:
            checksum = self.remainder
        else:
            checksum = self.minuend - self.remainder
        return str(self.reconcile(checksum))

    def adjust_input(self, account_code: str) -> str:
        return account_code

    def get_digits(self, account_code: str) -> str:
        positions = self.get_positions(account_code)
        # The positions are provided as in the specification, which starts counting at 1
        start, end = positions.start - 1, positions.end

        assert len(account_code) == ACCOUNT_CODE_LENGTH
        assert start >= 0 and start <= ACCOUNT_CODE_LENGTH
        assert end >= start and end <= ACCOUNT_CODE_LENGTH

        digits = account_code[start:end]
        if self.reverse:
            digits = digits[::-1]
        return digits

    def get_positions(self, account_code: str) -> Positions:
        return self.positions

    def compute_weighted_sum(self, digits: str) -> int:
        return sum(self.compute_summand(int(d), w) for d, w in zip(digits, cycle(self.weights)))

    def compute_summand(self, digit: int, weight: int) -> int:
        return digit * weight

    def compute_remainder(self, number: int) -> int:
        return number % self.modulus

    def reconcile(self, checksum: int) -> int:
        return 0 if checksum >= 10 else checksum

    def validate(self, account_code: str) -> bool:
        account_code = self.adjust_input(account_code)
        check_digit = self.compute(account_code)
        positions = self.get_positions(account_code)
        return check_digit == account_code[positions.check_digit - 1]


class WeightedMod10(WeightedModulus):
    modulus: int = 10
    minuend: Optional[int] = 10


class WeightedMod11(WeightedModulus):
    modulus: int = 11
    minuend: Optional[int] = 11


@register
class Algorithm00(WeightedMod10):
    name = "00"
    positions = Positions(start=1, end=9, check_digit=10)
    weights = [2, 1]

    def compute_summand(self, digit: int, weight: int) -> int:
        return digit_sum(super().compute_summand(digit, weight))


@register
class Algorithm01(WeightedMod10):
    name = "01"
    positions = Positions(start=1, end=9, check_digit=10)
    weights = [3, 7, 1]


@register
class Algorithm02(WeightedMod11):
    name = "02"
    positions = Positions(start=1, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7, 8, 9]

    def reconcile(self, checksum: int) -> int:
        if self.remainder == 0:
            return 0
        if self.remainder == 1:
            raise InvalidBBANChecksum(f"Invalid remaidner: {self.remainder}")
        return checksum


@register
class Algorithm03(Algorithm01):
    name = "03"
    weights = [2, 1]


@register
class Algorithm04(Algorithm02):
    name = "04"
    weights = [2, 3, 4, 5, 6, 7]


@register
class Algorithm05(Algorithm01):
    name = "05"
    weights = [7, 3, 1]


@register
class Algorithm06(WeightedMod11):
    name = "06"
    positions = Positions(start=1, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7]


@register
class Algorithm07(Algorithm02):
    name = "07"
    weights = [2, 3, 4, 5, 6, 7, 8, 9, 10]


@register
class Algorithm08(Algorithm00):
    name = "08"
    min_account_code = 6000

    def compute(self, account_code: str) -> str:
        if int(account_code) < self.min_account_code:
            return ""
        return super().compute(account_code)

    def validate(self, account_code: str) -> bool:
        if int(account_code) < self.min_account_code:
            return True
        return super().validate(account_code)


@register
class Algorithm09(checksum.Algorithm):
    name = "09"

    def compute(self, account_code: str) -> str:
        return ""

    def validate(self, account_code: str) -> bool:
        return True


@register
class Algorithm10(Algorithm06):
    name = "10"
    weights = [2, 3, 4, 5, 6, 7, 8, 9, 10]


@register
class Algorithm11(Algorithm10):
    name = "11"

    def reconcile(self, checksum: int) -> int:
        if checksum == 10:
            return 9
        return checksum


@register
class Algorithm13(Algorithm00):
    name = "13"
    positions = Positions(start=2, end=7, check_digit=8)
    weights = [2, 1]


@register
class Algorithm14(Algorithm02):
    name = "14"
    positions = Positions(start=4, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7]


@register
class Algorithm15(Algorithm06):
    name = "15"
    positions = Positions(start=6, end=9, check_digit=10)
    weights = [2, 3, 4, 5]


@register
class Algorithm16(Algorithm06):
    name = "16"
    positions = Positions(start=6, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7]

    def validate(self, account_code: str) -> bool:
        check_digit = self.compute(account_code)
        if self.remainder == 1 and account_code[8] == account_code[9]:
            return True
        return check_digit == account_code[self.positions.check_digit - 1]


@register
class Algorithm17(WeightedMod11):
    name = "17"
    minuend = 10
    positions = Positions(start=2, end=7, check_digit=8)
    reverse = False
    weights = [1, 2]

    def compute_weighted_sum(self, digits: str) -> int:
        return super().compute_weighted_sum(digits) - 1

    def compute_summand(self, digit: int, weight: int) -> int:
        return digit_sum(super().compute_summand(digit, weight))


@register
class Algorithm18(Algorithm01):
    name = "18"
    weights = [3, 9, 7, 1]


@register
class Algorithm19(Algorithm06):
    name = "19"
    weights = [2, 3, 4, 5, 6, 7, 8, 9, 1]


@register
class Algorithm20(Algorithm06):
    name = "20"
    weights = [2, 3, 4, 5, 6, 7, 8, 9, 3]


@register
class Algorithm21(Algorithm00):
    name = "21"
    weights = [2, 1]

    def compute_remainder(self, number: int) -> int:
        while number >= 10:
            number = digit_sum(number)
        return number


@register
class Algorithm22(Algorithm01):
    name = "22"
    weights = [3, 1]

    def compute_summand(self, digit: int, weight: int) -> int:
        return super().compute_summand(digit, weight) % 10


@register
class Algorithm23(Algorithm16):
    name = "23"
    positions = Positions(start=1, end=6, check_digit=7)
    weights = [2, 3, 4, 5, 6, 7]


@register
class Algorithm24(WeightedMod10):
    name = "24"
    minuend = None
    positions = Positions(start=1, end=9, check_digit=10)
    reverse = False
    weights = [1, 2, 3]

    def get_digits(self, account_code: str) -> str:
        digits = super().get_digits(account_code)
        if int(digits[0]) in {3, 4, 5, 6}:
            digits = digits[1:]
        elif int(digits[0]) == 9:
            digits = digits[3:]
        return digits.lstrip("0")

    def compute_summand(self, digit: int, weight: int) -> int:
        return (super().compute_summand(digit, weight) + weight) % 11


@register
class Algorithm25(WeightedMod11):
    name = "25"
    positions = Positions(start=2, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7, 8, 9]

    def validate(self, account_code: str) -> bool:
        result = super().validate(account_code)
        if self.remainder == 1 and account_code[1] not in {"8", "9"}:
            return False
        return result


@register
class Algorithm26(Algorithm06):
    name = "26"
    positions = Positions(start=1, end=7, check_digit=8)
    weights = [2, 3, 4, 5, 6, 7]

    def adjust_input(self, account_code: str) -> str:
        if account_code.startswith("00"):
            account_code = account_code[2:] + "00"
        return account_code


@register
class Algorithm28(Algorithm06):
    name = "28"
    positions = Positions(start=1, end=7, check_digit=8)
    weights = [2, 3, 4, 5, 6, 7, 8]


@register
class Algorithm32(Algorithm06):
    name = "32"
    positions = Positions(start=4, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7]


@register
class Algorithm33(Algorithm06):
    name = "33"
    positions = Positions(start=5, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6]


@register
class Algorithm34(Algorithm06):
    name = "34"
    positions = Positions(start=1, end=7, check_digit=8)
    weights = [2, 4, 8, 5, 10, 9, 7]


@register
class Algorithm38(Algorithm06):
    name = "38"
    positions = Positions(start=4, end=9, check_digit=10)
    weights = [2, 4, 8, 5, 10, 9]


@register
class Algorithm60(Algorithm00):
    name = "60"
    positions = Positions(start=3, end=9, check_digit=10)


@register
class Algorithm61(Algorithm00):
    name = "61"
    positions = Positions(start=1, end=7, check_digit=8)
    weights = [2, 1]

    def get_digits(self, account_code: str) -> str:
        digits = super().get_digits(account_code)
        if account_code[8] == "8":
            digits = account_code[:7:-1] + digits
        return digits


@register
class Algorithm63(WeightedMod10):
    name = "63"
    positions = Positions(start=2, end=7, check_digit=8)
    weights = [2, 1]

    def compute_summand(self, digit: int, weight: int) -> int:
        return digit_sum(super().compute_summand(digit, weight))

    def validate(self, account_code: str) -> bool:
        if account_code[0] != "0":
            return False
        return super().validate(account_code)


@register
class Algorithm68(Algorithm00):
    name = "68"
    positions = Positions(start=1, end=9, check_digit=10)
    weights = [2, 1]

    def get_digits(self, account_code: str) -> str:
        digits = super().get_digits(account_code)
        # The digits are already reversed at this point. The algorithm counts the positions from
        # right to left, so that the 7th position is at index 5, when we consider that the 1st
        # position has already been excluded because it is the check digit.
        digits = digits.rstrip("0")
        if len(digits) == 9:
            if digits[5] != "9":
                raise InvalidBBANChecksum(
                    "10 digit long acccount codes require the 7th digit to be set to 9"
                )
            digits = digits[:6]
        return digits

    def validate(self, account_code: str) -> bool:
        if 400_000_000 <= int(account_code) <= 499_999_999:
            return True
        if super().validate(account_code) is False:
            # If the checksum calculation fails, the algorithm should be executed again, with the
            # 7th and 8th position of the account code removed. Since the positions are counted from
            # right to left, they translate into indices 2 and 3 counting from left.
            check_digit = self.compute(account_code[:2] + "00" + account_code[4:])
            return check_digit == account_code[self.positions.check_digit - 1]
        return True


@register
class Algorithm76(WeightedMod11):
    name = "76"
    minuend = None
    positions = Positions(start=2, end=7, check_digit=8)
    weights = [2, 3, 4, 5, 6, 7, 8]

    def get_digits(self, account_code: str) -> str:
        digits = super().get_digits(account_code)
        return digits.rstrip("0")

    def validate(self, account_code: str) -> bool:
        if int(account_code[0]) not in {0, 4, 6, 7, 8, 9}:
            return False
        return super().validate(account_code)


@register
class Algorithm88(Algorithm06):
    name = "88"
    positions = Positions(start=4, end=9, check_digit=10)
    weights = [2, 3, 4, 5, 6, 7, 8]

    def get_positions(self, account_code: str) -> Positions:
        if account_code[2] == "9":
            return Positions(start=3, end=9, check_digit=10)
        return super().get_positions(account_code)


@register
class Algorithm91(checksum.Algorithm):
    name = "91"

    class Variant1(Algorithm06):
        positions = Positions(start=1, end=6, check_digit=7)
        weights = [2, 3, 4, 5, 6, 7]

    class Variant2(Variant1):
        weights = [7, 6, 5, 4, 3, 2]

    class Variant3(Variant1):
        positions = Positions(start=1, end=10, check_digit=7)
        weights = [2, 3, 4, 0, 5, 6, 7, 8, 9, 10]

    class Variant4(Variant1):
        weights = [2, 4, 8, 5, 10, 9]

    def compute(self, account_code: str) -> str:
        return self.Variant1().compute(account_code)

    def validate(self, account_code: str) -> bool:
        for algo_cls in [self.Variant1, self.Variant2, self.Variant3, self.Variant4]:
            if algo_cls().validate(account_code):
                return True
        return False


@register
class Algorithm99(Algorithm06):
    name = "99"

    def validate(self, account_code: str) -> bool:
        if account_code in {"0499999999", "0396000000"}:
            return True
        return super().validate(account_code)



import re
from functools import total_ordering
from typing import Any
from typing import Optional


_clean_regex = re.compile(r"\s+")


@total_ordering
class Base:
    def __init__(self, code: str) -> None:
        self._code = clean(code)

    def __str__(self) -> str:
        return self.compact

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}={self!s}>"

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)

    def __lt__(self, other: Any) -> bool:
        return str(self) < str(other)

    def __len__(self) -> int:
        return self.length

    @property
    def compact(self) -> str:
        """str: Compact representation of the code."""
        return self._code

    @property
    def length(self) -> int:
        """int: Length of the compact code."""
        return len(self.compact)

    def _get_component(self, start: int, end: Optional[int] = None) -> str:
        if start < self.length and (end is None or end <= self.length):
            return self.compact[start:end] if end else self.compact[start:]
        return ""


def clean(s: str) -> str:
    return _clean_regex.sub("", s).upper()



import string
from functools import partial

from schwifty import checksum
from schwifty import registry


register = partial(checksum.register, prefix="IT")


def get_index(char: str) -> int:
    try:
        return string.digits.index(char)
    except ValueError:
        return string.ascii_uppercase.index(char.upper())


@register
class DefaultAlgorithm(checksum.Algorithm):
    name = "default"
    accepts = checksum.InputType.BBAN
    checksum_length = 1

    def compute(self, bban: str) -> str:
        spec = registry.get("iban")
        assert isinstance(spec, dict)

        odds = (
            [1, 0, 5, 7, 9, 13, 15]
            + [17, 19, 21, 2, 4, 18, 20]
            + [11, 3, 6, 8, 12, 14, 16]
            + [10, 22, 25, 24, 23]
        )
        sum_ = 0
        for i, char in enumerate(bban[self.checksum_length - spec["IT"]["bban_length"] :]):
            if (i + 1) % 2 == 0:
                sum_ += get_index(char)
            else:
                sum_ += odds[get_index(char)]
        return string.ascii_uppercase[sum_ % 26]

    def validate(self, bban: str) -> bool:
        return bban[0] == self.compute(bban[self.checksum_length :])



import csv
import json

import requests

from schwifty import BIC


URL = "https://www.nbs.sk/_img/Documents/_PlatobneSystemy/EUROSIPS/Directory_IC_DPS_SR.csv"


def process():
    with requests.get(URL, stream=True) as fp:
        csvfile = csv.reader([line.decode("cp1250") for line in fp.iter_lines()], delimiter=";")
    registry = []
    next(csvfile)
    for row in csvfile:
        if len(row) != 4:
            continue
        bic = row[2].strip().upper()
        if bic and not BIC(bic, allow_invalid=True).is_valid:
            continue
        registry.append(
            {
                "country_code": "SK",
                "primary": True,
                "bic": bic,
                "bank_code": row[0].strip().zfill(4),
                "name": row[1].strip(),
                "short_name": row[1].strip(),
            }
        )
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_sk.json", "w") as fp:
        json.dump(process(), fp, indent=2)



#!/usr/bin/env python
import json

import requests
import xlrd


URL = "https://www.betaalvereniging.nl/wp-content/uploads/BIC-lijst-NL.xlsx"


def process():
    registry = []

    book = xlrd.open_workbook(file_contents=requests.get(URL).content)
    sheet = book.sheet_by_index(0)

    for row in list(sheet.get_rows())[4:]:
        bic, bank_code, name = row[:3]

        if not any((bic.value, bank_code.value, name.value)):
            continue

        registry.append(
            {
                "country_code": "NL",
                "primary": True,
                "bic": bic.value.upper(),
                "bank_code": bank_code.value,
                "name": name.value,
                "short_name": name.value,
            }
        )
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_nl.json", "w") as fp:
        json.dump(process(), fp, indent=2)



#!/usr/bin/env python
import json
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup


FIELD_LENGTHS = {
    "bank_code": 8,
    "feature": 1,
    "name": 58,
    "postal_code": 5,
    "place": 35,
    "short_name": 27,
    "pan": 5,
    "bic": 11,
    "check_digit_method": 2,
    "record_number": 6,
    "mod_number": 1,
    "tbd": 1,
    "successor_bank_code": 8,
}

URL = "https://www.bundesbank.de/de/aufgaben/unbarer-zahlungsverkehr/serviceangebot/bankleitzahlen"


def get_download_url():
    soup = BeautifulSoup(requests.get(URL).content, "html.parser")
    atag = soup.find(href=lambda ref: ref and "download-bankleitzahlen" in ref)

    soup = BeautifulSoup(requests.get(urljoin(URL, atag.get("href"))).content, "html.parser")
    atag = soup.find(href=lambda ref: ref and "blz-aktuell-txt-data.txt" in ref)
    return urljoin(URL, atag.get("href"))


def get_raw():
    return requests.get(get_download_url()).content.decode(encoding="latin1")


def parse(raw):
    for line in raw.split("\n"):
        if not line:
            continue
        record = {}
        offset = 0
        for field, length in FIELD_LENGTHS.items():
            record[field] = line[offset : offset + length].strip()
            offset = offset + length
        yield record


def process(records):
    fieldnames = ("bank_code", "name", "short_name", "bic")
    registry = []
    for record in records:
        if not record["bank_code"]:
            continue

        cleaned = {k: v for k, v in record.items() if k in fieldnames}
        cleaned["primary"] = record["feature"] == "1"
        cleaned["country_code"] = "DE"
        cleaned["checksum_algo"] = record["check_digit_method"]

        if cleaned["bic"]:
            registry.append(cleaned)
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_de.json", "w") as fp:
        json.dump(process(parse(get_raw())), fp, indent=2)



import json
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup


BASE_URLS = [
    "https://www.iban.es/bancos/",
    "https://www.iban.es/bancos-extranjeros/",
    "https://www.iban.es/cajas/",
]
titles = {
    "Código de entidad (código de banco)": "bank_code",
    "Código BIC Banco (Codigo SWIFT)": "bic",
    "Denominación legal del Banco": "name",
    "Nombre Comercial (abreviado)": "short_name",
}


def get_bank_details(url):
    soup = BeautifulSoup(requests.get(url).content, "html.parser")
    rows = soup.select("div.about-content-text table tr")

    record = {"country_code": "ES", "primary": True}
    for row in rows:
        cells = row.find_all("td")
        if len(cells) != 2:
            continue
        title, value = cells[0].text.strip(), cells[1].text.strip()
        key = titles.get(title)
        if key:
            record[key] = value
    return record


def process():
    result = []
    for url in BASE_URLS:
        soup = BeautifulSoup(requests.get(url).content, "html.parser")
        paths = [str(a["href"]) for a in soup.select("h6.portfolio-title a")]
        print(f"Fetched {len(paths)} bank records")
        result.extend([get_bank_details(urljoin(url, path)) for path in paths])
    return result


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_es.json", "w") as fp:
        json.dump(process(), fp, indent=2)



import json

import requests


URL = (
    "https://www.bsi.si/ckfinder/connector?command=Proxy&lang=sl&type=Files&"
    "currentFolder=%2FPla%C4%8Dilni%20sistemi%2FSeznam%20identifikacijskih%20oznak%20PPS%2F&"
    "hash=6ce6c512ea433a7fc5c8841628e7696cd0ff7f2b&fileName=List-of-identifiers-of-PSP.txt"
)


def process():
    registry = []
    with requests.get(URL, stream=True) as txtfile:
        for row in list(txtfile.iter_lines())[1:]:
            cells = [cell.strip() for cell in row.decode("latin1").split(";")]
            if len(cells) != 6:
                continue
            bank_code, name = cells[:2]
            registry.append(
                {
                    "country_code": "SI",
                    "primary": True,
                    "bic": cells[5].upper(),
                    "bank_code": bank_code,
                    "name": name,
                    "short_name": name,
                }
            )
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_si.json", "w") as fp:
        json.dump(process(), fp, indent=2)



import abc
import enum
from typing import ClassVar
from typing import Dict
from typing import Optional
from typing import Type


class InputType(enum.Enum):
    BBAN = enum.auto()
    ACCOUNT_CODE = enum.auto()


class Algorithm(metaclass=abc.ABCMeta):
    name: ClassVar[str]
    accepts: ClassVar[InputType] = InputType.ACCOUNT_CODE

    @abc.abstractmethod
    def compute(self, account_code: str) -> str:
        pass

    @abc.abstractmethod
    def validate(self, account_code: str) -> bool:
        pass


algorithms: Dict[str, Algorithm] = {}


def register(algorithm_cls: Type[Algorithm], prefix: Optional[str] = None) -> Type[Algorithm]:
    key = algorithm_cls.name
    if prefix is not None:
        key = f"{prefix}:{key}"
    algorithms[key] = algorithm_cls()
    return algorithm_cls


from schwifty.checksum import belgium  # noqa
from schwifty.checksum import germany  # noqa
from schwifty.checksum import italy  # noqa
from schwifty.checksum import france  # noqa



class SchwiftyException(ValueError):
    """Base exception of all schwifty related errors."""


class InvalidLength(SchwiftyException):
    """Indicates that the length of the input does not match the specifcation."""


class InvalidStructure(SchwiftyException):
    """Indicates a strctural error of the input (e.g. invalid characters)."""


class InvalidCountryCode(SchwiftyException):
    """Unknown country code in the input."""


class InvalidBankCode(SchwiftyException):
    """Indicates that the bank code has an invalid structure."""


class InvalidBranchCode(SchwiftyException):
    """Indicates that the branch code has an invalid strucutre."""


class InvalidAccountCode(SchwiftyException):
    """Indicates that the account code has an invalid strucutre."""


class InvalidChecksumDigits(SchwiftyException):
    """Indicates that the IBAN's checksum is invalid."""


class InvalidBBANChecksum(SchwiftyException):
    """Indicates that the BBAN's checksum is invalid."""



import functools

from schwifty import checksum
from schwifty import registry


register = functools.partial(checksum.register, prefix="FR")

numerics = {
    "0": "0",
    "1": "1",
    "2": "2",
    "3": "3",
    "4": "4",
    "5": "5",
    "6": "6",
    "7": "7",
    "8": "8",
    "9": "9",
    "A": "1",
    "B": "2",
    "C": "3",
    "D": "4",
    "E": "5",
    "F": "6",
    "G": "7",
    "H": "8",
    "I": "9",
    "J": "1",
    "K": "2",
    "L": "3",
    "M": "4",
    "N": "5",
    "O": "6",
    "P": "7",
    "Q": "8",
    "R": "9",
    "S": "2",
    "T": "3",
    "U": "4",
    "V": "5",
    "W": "6",
    "X": "7",
    "Y": "8",
    "Z": "9",
}


def convert(code: str) -> int:
    res = ""
    for c in code:
        res += numerics[c]
    return int(res)


@register
class DefaultAlgorithm(checksum.Algorithm):
    name = "default"
    accepts = checksum.InputType.BBAN
    checksum_length = 2

    def compute(self, bban: str) -> str:
        spec = registry.get("iban")
        assert isinstance(spec, dict)

        positions = spec["FR"]["positions"]
        bank_code = bban[slice(*positions["bank_code"])]
        branch_code = bban[slice(*positions["branch_code"])]
        account_code = bban[slice(*positions["account_code"])]

        checksum = 97 - (
            (89 * convert(bank_code) + 15 * convert(branch_code) + 3 * convert(account_code)) % 97
        )
        return str(checksum).zfill(self.checksum_length)

    def validate(self, bban: str) -> bool:
        return bban[-self.checksum_length :] == self.compute(bban[: -self.checksum_length])



#!/usr/bin/env python
import json

import requests
import xlrd


URL = "https://www.nbb.be/doc/be/be/protocol/r_fulllist_of_codes_current.xlsx"


def process():
    registry = []
    skip_names = ["NAV", "VRIJ", "NAP", "NYA", "VRIJ - LIBRE", "-"]

    book = xlrd.open_workbook(file_contents=requests.get(URL).content)
    sheet = book.sheet_by_index(0)

    for row in list(sheet.get_rows())[2:]:
        bank_code, bic, name, second_name = row[:4]
        if bic.value.upper() in skip_names:
            continue
        registry.append(
            {
                "country_code": "BE",
                "primary": True,
                "bic": bic.value.upper().replace(" ", ""),
                "bank_code": bank_code.value,
                "name": name.value or second_name.value,
                "short_name": name.value or second_name.value,
            }
        )
    return registry


if __name__ == "__main__":
    with open("schwifty/bank_registry/generated_be.json", "w") as fp:
        json.dump(process(), fp, indent=2)



import itertools
import json
import pathlib
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union


try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files  # type: ignore


_registry: Dict[str, Union[Dict, List[Dict]]] = {}


def merge_dicts(left: Dict, right: Dict) -> Dict:
    merged = {}
    for key in frozenset(right) & frozenset(left):
        left_value, right_value = left[key], right[key]
        if isinstance(left_value, Dict) and isinstance(right_value, Dict):
            merged[key] = merge_dicts(left_value, right_value)
        else:
            merged[key] = right_value

    for key, value in itertools.chain(left.items(), right.items()):
        if key not in merged:
            merged[key] = value
    return merged


def has(name: str) -> bool:
    return name in _registry


def get(name: str) -> Union[Dict, List[Dict]]:
    if not has(name):
        data = None
        directory = files(__package__) / f"{name}_registry"
        assert isinstance(directory, pathlib.Path)
        for entry in sorted(directory.glob("*.json")):
            with entry.open(encoding="utf-8") as fp:
                chunk = json.load(fp)
                if data is None:
                    data = chunk
                elif isinstance(data, list):
                    data.extend(chunk)
                elif isinstance(data, dict):
                    data = merge_dicts(data, chunk)
        if data is None:
            raise ValueError(f"Failed to load registry {name}")
        save(name, data)
    return _registry[name]


def save(name: str, data: Union[Dict, List[Dict]]) -> None:
    _registry[name] = data


def build_index(
    base_name: str,
    index_name: str,
    key: Union[str, Tuple],
    accumulate: bool = False,
    **predicate: Any,
) -> None:
    def make_key(entry: Dict) -> Union[Tuple, str]:
        return tuple(entry[k] for k in key) if isinstance(key, tuple) else entry[key]

    def match(entry: dict) -> bool:
        return all(entry[key] == value for key, value in predicate.items())

    base = get(base_name)
    assert isinstance(base, list)
    if accumulate:
        data = defaultdict(list)
        for entry in base:
            if not match(entry):
                continue
            data[make_key(entry)].append(entry)
        save(index_name, dict(data))
    else:
        save(index_name, {make_key(entry): entry for entry in base if match(entry)})


def manipulate(name: str, func: Callable) -> None:
    registry = get(name)
    if isinstance(registry, dict):
        for key, value in registry.items():
            registry[key] = func(key, value)
    elif isinstance(registry, list):
        registry = [func(item) for item in registry]
    save(name, registry)



from __future__ import annotations

import re
import string
from typing import Dict
from typing import Optional

from pycountry import countries
from pycountry.db import Data

from schwifty import common
from schwifty import exceptions
from schwifty import registry
from schwifty.bic import BIC
from schwifty.checksum import algorithms
from schwifty.checksum import InputType


_spec_to_re: Dict[str, str] = {"n": r"\d", "a": r"[A-Z]", "c": r"[A-Za-z0-9]", "e": r" "}

_alphabet: str = string.digits + string.ascii_uppercase


def _get_iban_spec(country_code: str) -> dict:
    try:
        spec = registry.get("iban")
        assert isinstance(spec, dict)
        return spec[country_code]
    except KeyError:
        raise exceptions.InvalidCountryCode(f"Unknown country-code '{country_code}'")


def numerify(string: str) -> int:
    return int("".join(str(_alphabet.index(c)) for c in string))


def code_length(spec: Dict, code_type: str) -> int:
    start, end = spec["positions"][code_type]
    return end - start


def add_bban_checksum(country_code: str, bban: str) -> str:
    if country_code in ["IT", "SM"]:
        # The IBAN of San Marino is covered by the Italian IBAN and uses the same checksum
        checksum = algorithms["IT:default"].compute(bban[1:])
        bban = checksum + bban[1:]
    elif country_code == "BE":
        # The Belgian BBAN format is XXXYYYYYYYZZ where:
        # - XXX: bank code
        # - YYYYYYY: account number
        # - ZZ: mod 97 remainder of XXYYYYYYY
        # The bban passed to this function has this format XXX00YYYYY
        bban = bban[:3] + bban[5:]
        checksum = algorithms["BE:default"].compute(bban)
        bban = bban + checksum
    elif country_code == "FR":
        checksum = algorithms["FR:default"].compute(bban)
        bban = bban[0:21] + checksum
    return bban


class IBAN(common.Base):
    """The IBAN object.

    Examples:

        You create a new IBAN object by supplying an IBAN code in text form. The IBAN
        is validated behind the scenes and you can then access all relevant components
        as properties::

            >>> iban = IBAN('DE89 3704 0044 0532 0130 00')
            >>> iban.account_code
            '0532013000'
            >>> iban.bank_code
            '37040044'
            >>> iban.country_code
            'DE'
            >>> iban.checksum_digits
            '89'


    Args:
        iban (str): The IBAN code.
        allow_invalid (bool): If set to `True` IBAN validation is skipped on instantiation.
        validate_bban (bool): If set to `True` also check the country specific checksum with then
                              BBAN.

    Raises:
        InvalidStructure: If the IBAN contains invalid characters or the BBAN does not match the
                          country specific format.
        InvalidChecksumDigits: If the IBAN's checksum is invalid.
        InvalidLength: If the length does not match the country specific specification.

    .. versionchanged:: 2021.05.1
        Added the `validate_bban` parameter that controls if the country specific checksum within
        the BBAN is also validated.
    """

    def __init__(self, iban: str, allow_invalid: bool = False, validate_bban: bool = False) -> None:
        super().__init__(iban)
        if self.checksum_digits == "??":
            self._code = self.country_code + self._calc_checksum_digits() + self.bban

        if not allow_invalid:
            self.validate(validate_bban)

    def _calc_checksum_digits(self) -> str:
        return "{:02d}".format(98 - (numerify(self.bban + self.country_code) * 100) % 97)

    @classmethod
    def generate(
        cls, country_code: str, bank_code: str, account_code: str, branch_code: str = ""
    ) -> IBAN:
        """Generate an IBAN from it's components.

        If the bank-code and/or account-number have less digits than required by their
        country specific representation, the respective component is padded with zeros.

        Examples:

            To generate an IBAN do the following::

                >>> bank_code = '37040044'
                >>> account_code = '532013000'
                >>> iban = IBAN.generate('DE', bank_code, account_code)
                >>> iban.formatted
                'DE89 3704 0044 0532 0130 00'

        Args:
            country_code (str): The ISO 3166 alpha-2 country code.
            bank_code (str): The country specific bank-code.
            account_code (str): The customer specific account-code.

        .. versionchanged:: 2020.08.3
            Added the `branch_code` parameter to allow the branch code (or sort code) to be
            specified independently.

        .. versionchanged:: 2021.05.2
            Added support for generating the country specific checksum of the BBAN for Belgian
            banks.
        """
        spec: Dict = _get_iban_spec(country_code)
        bank_code_length: int = code_length(spec, "bank_code")
        branch_code_length: int = code_length(spec, "branch_code")
        account_code_length: int = code_length(spec, "account_code")

        if len(bank_code) == bank_code_length + branch_code_length:
            bank_code, branch_code = bank_code[:bank_code_length], bank_code[bank_code_length:]

        if len(bank_code) > bank_code_length:
            raise exceptions.InvalidBankCode(f"Bank code exceeds maximum size {bank_code_length}")

        if len(branch_code) > branch_code_length:
            raise exceptions.InvalidBranchCode(
                f"Branch code exceeds maximum size {branch_code_length}"
            )

        if len(account_code) > account_code_length:
            raise exceptions.InvalidAccountCode(
                f"Account code exceeds maximum size {account_code_length}"
            )

        bban = "0" * spec["bban_length"]
        positions = spec["positions"]
        components = {
            "bank_code": bank_code,
            "branch_code": branch_code,
            "account_code": account_code,
        }
        for key, value in components.items():
            end = positions[key][1]
            start = end - len(value)
            bban = bban[:start] + value + bban[end:]

        bban = add_bban_checksum(country_code, bban)
        return cls(country_code + "??" + bban)

    def validate(self, validate_bban: bool = False) -> bool:
        """Validate the structural integrity of this IBAN.

        This function will verify the country specific format as well as the Luhn checksum in the
        3rd and 4th position of the IBAN. For some countries (currently Germany and Italy) it will
        also verify the correctness of the country specific checksum within the BBAN if the
        `validate_bban` parameter is set to `True`. For German banks it will pick the appropriate
        algorithm based on the bank code and verify that the account code has the correct checksum.

        Note:
            You have to use the `allow_invalid` paramter when constructing the :class:`IBAN`-object
            to circumvent the implicit validation.

        Raises:
            InvalidStructure: If the IBAN contains invalid characters or the BBAN does not match the
                              country specific format.
            InvalidChecksumDigits: If the IBAN's checksum is invalid.
            InvalidLength: If the length does not match the country specific specification.

        .. versionchanged:: 2021.05.1
            Added the `validate_bban` parameter that controls if the country specific checksum
            within the BBAN is also validated.
        """
        self._validate_characters()
        self._validate_length()
        self._validate_format()
        self._validate_iban_checksum()
        if validate_bban:
            self._validate_bban_checksum()
        return True

    def _validate_characters(self) -> None:
        if not re.match(r"[A-Z]{2}\d{2}[A-Z]*", self.compact):
            raise exceptions.InvalidStructure(f"Invalid characters in IBAN {self.compact}")

    def _validate_length(self) -> None:
        if self.spec["iban_length"] != self.length:
            raise exceptions.InvalidLength("Invalid IBAN length")

    def _validate_format(self) -> None:
        if not self.spec["regex"].match(self.bban):
            raise exceptions.InvalidStructure(
                "Invalid BBAN structure: '{}' doesn't match '{}''".format(
                    self.bban, self.spec["bban_spec"]
                )
            )

    def _validate_iban_checksum(self) -> None:
        if self.numeric % 97 != 1 or self._calc_checksum_digits() != self.checksum_digits:
            raise exceptions.InvalidChecksumDigits("Invalid checksum digits")

    def _validate_bban_checksum(self) -> None:
        bank = self.bank or {}
        algo_name = bank.get("checksum_algo", "default")
        algo = algorithms.get(f"{self.country_code}:{algo_name}")
        if algo is None:
            return
        if algo.accepts == InputType.ACCOUNT_CODE:
            value = self.account_code
        elif algo.accepts == InputType.BBAN:
            value = self.bban
        else:
            assert False, "Unsupported checksum algorithm input type"
        if not algo.validate(value):
            raise exceptions.InvalidBBANChecksum("Invalid BBAN checksum")

    @property
    def is_valid(self) -> bool:
        """bool: Indicate if this is a valid IBAN.

        Note:
            You have to use the `allow_invalid` paramter when constructing the :class:`IBAN`-object
            to circumvent the implicit validation.

        Examples:
            >>> IBAN('AB1234567890', allow_invalid=True).is_valid
            False

        .. versionadded:: 2020.08.1
        """
        try:
            return self.validate()
        except exceptions.SchwiftyException:
            return False

    @property
    def numeric(self) -> int:
        """int: A numeric represenation of the IBAN."""
        return numerify(self.bban + self.compact[:4])

    @property
    def formatted(self) -> str:
        """str: The IBAN formatted in blocks of 4 digits."""
        return " ".join(self.compact[i : i + 4] for i in range(0, len(self.compact), 4))

    @property
    def spec(self) -> Dict:
        """dict: The country specific IBAN specification."""
        return _get_iban_spec(self.country_code)

    @property
    def bic(self) -> Optional[BIC]:
        """BIC or None: The BIC associated to the IBAN´s bank-code.

        If the bank code is not available in Schwifty's registry ``None`` is returned.

        .. versionchanged:: 2020.08.1
            Returns ``None`` if no appropriate :class:`BIC` can be constructed.
        """
        try:
            return BIC.from_bank_code(self.country_code, self.bank_code or self.branch_code)
        except exceptions.SchwiftyException:
            return None

    @property
    def country(self) -> Optional[Data]:
        """Country: The country this IBAN is registered in."""
        return countries.get(alpha_2=self.country_code)

    def _get_code(self, code_type: str) -> str:
        start, end = self.spec["positions"][code_type]
        return self.bban[start:end]

    @property
    def bban(self) -> str:
        """str: The BBAN part of the IBAN."""
        return self._get_component(start=4)

    @property
    def country_code(self) -> str:
        """str: ISO 3166 alpha-2 country code."""
        return self._get_component(start=0, end=2)

    @property
    def checksum_digits(self) -> str:
        """str: Two digit checksum of the IBAN."""
        return self._get_component(start=2, end=4)

    @property
    def bank_code(self) -> str:
        """str: The country specific bank-code."""
        return self._get_code(code_type="bank_code")

    @property
    def branch_code(self) -> str:
        """str or None: The branch-code of the bank if available."""
        return self._get_code(code_type="branch_code")

    @property
    def account_code(self) -> str:
        """str: The customer specific account-code"""
        return self._get_code(code_type="account_code")

    @property
    def bank(self) -> Optional[dict]:
        bank_registry = registry.get("bank_code")
        assert isinstance(bank_registry, dict)
        return bank_registry.get((self.country_code, self.bank_code))

    @property
    def bank_name(self) -> Optional[str]:
        """Optional[str]: The name of the bank associated with the IBAN bank code.

        Examples:
            >>> IBAN('DE89370400440532013000').bank_name
            'Commerzbank'

        .. versionadded:: 2022.04.2
        """

        return None if self.bank is None else self.bank["name"]

    @property
    def bank_short_name(self) -> Optional[str]:
        """Optional[str]: The name of the bank associated with the IBAN bank code.

        Examples:
            >>> IBAN('DE89370400440532013000').bank_short_name
            'Commerzbank Köln'

        .. versionadded:: 2022.04.2
        """

        return None if self.bank is None else self.bank["short_name"]


def add_bban_regex(country: str, spec: Dict) -> Dict:
    bban_spec = spec["bban_spec"]
    spec_re = r"(\d+)(!)?([{}])".format("".join(_spec_to_re.keys()))

    def convert(match: re.Match) -> str:
        quantifier = ("{%s}" if match.group(2) else "{1,%s}") % match.group(1)
        return _spec_to_re[match.group(3)] + quantifier

    spec["regex"] = re.compile("^{}$".format(re.sub(spec_re, convert, bban_spec)))
    return spec


registry.manipulate("iban", add_bban_regex)



import functools

from schwifty import checksum
from schwifty import registry


register = functools.partial(checksum.register, prefix="BE")


@register
class DefaultAlgorithm(checksum.Algorithm):
    name = "default"
    accepts = checksum.InputType.BBAN
    checksum_length = 2

    def compute(self, bban: str) -> str:
        spec = registry.get("iban")
        assert isinstance(spec, dict)
        checksum = int(bban[: spec["BE"]["bban_length"] - self.checksum_length]) % 97
        return str(checksum).zfill(self.checksum_length)

    def validate(self, bban: str) -> bool:
        return bban[-self.checksum_length :] == self.compute(bban[: -self.checksum_length])
